from typing import List
from showml.utils.dataset import Dataset
import pytest

from showml.utils.exceptions import DataTypeError, DatasetSizeMismatchError
import numpy as np
from numpy.testing import assert_almost_equal


def test_string_input() -> None:
    with pytest.raises(DataTypeError):
        ds = Dataset("X", "Y")


def test_list_input() -> None:
    X = [8, 307, 130, 3504, 12, 70, 1]
    y = [15]
    with pytest.raises(DataTypeError):
        ds = Dataset(X, y)


def test_multi_datatype_list_input() -> None:
    X = ["8", 307.998, "130", 3504.9999, 12, 70, 1]
    y = [15]
    with pytest.raises(DataTypeError):
        ds = Dataset(X, y)


def test_empty_input() -> None:
    X: List[None] = []
    y = [15]
    with pytest.raises(DataTypeError):
        ds = Dataset(X, y)


def test_single_sample_input() -> None:
    X = np.array([[8, 307, 130, 3504, 12, 70, 1]])
    y = np.array([[15]])
    ds = Dataset(X, y)
    assert_almost_equal(ds.X, np.array([[8, 307, 130, 3504, 12, 70, 1]]))
    assert_almost_equal(ds.y, np.array([[15]]))


def test_size_X_mismatch() -> None:
    X = np.array(
        [[8, 454, 220, 4354, 9], [9, 442, 21, 4354, 9], [10, 444, 220, 4354, 10]]
    )
    y = np.array([14])
    with pytest.raises(DatasetSizeMismatchError):
        ds = Dataset(X, y)


def test_size_y_mismatch() -> None:
    X = np.array([[8, 454, 220, 4354, 9]])
    y = np.array([[14], [20], [21]])
    with pytest.raises(DatasetSizeMismatchError):
        ds = Dataset(X, y)


def test_size_mismatch_X_empty() -> None:
    X = np.array([[]])
    y = np.array([[14], [15]])
    with pytest.raises(DatasetSizeMismatchError):
        ds = Dataset(X, y)


def test_size_mismatch_y_empty() -> None:
    X = np.array(
        [[8, 454, 220, 4354, 9], [9, 442, 21, 4354, 9], [10, 444, 220, 4354, 10]]
    )
    y = np.array([[]])
    with pytest.raises(DatasetSizeMismatchError):
        ds = Dataset(X, y)


def test_correct_multisample_input() -> None:
    X = np.array(
        [[8, 454, 220, 4354, 9], [9, 442, 21, 4354, 9], [10, 444, 220, 4354, 10]]
    )
    y = np.array([[14], [20], [21]])
    ds = Dataset(X, y)
    assert_almost_equal(
        ds.X,
        np.array(
            [[8, 454, 220, 4354, 9], [9, 442, 21, 4354, 9], [10, 444, 220, 4354, 10]]
        ),
    )
    assert_almost_equal(ds.y, np.array([[14], [20], [21]]))



from enum import Enum, auto
from dataclasses import dataclass


class Action(Enum):
    """This class defines the different Actions in the game.
    """

    START = auto()
    STOP = auto()
    RESET = auto()
    CELL_TOGGLE = auto()
    NO_EVENT = auto()


@dataclass
class Event:
    """This class defines an Event (and also the row, column if a cell is toggled)
    """

    action: Action
    row: int = 0
    column: int = 0



class DatasetSizeMismatchError(Exception):
    """Raised when X and y have different lengths.
    """


class DataTypeError(Exception):
    """Raised when a variable has an unexpected datatype.
    """


class InvalidShapeError(Exception):
    """Raised when an unexpected array shape is encountered.
    """


class InvalidValueError(Exception):
    """Raised when an unexpected value is encountered.
    """



from typing import Generator, Tuple

import numpy as np

from showml.utils.exceptions import InvalidShapeError


def initialize_params(num_dimensions: int) -> Tuple[np.ndarray, np.ndarray]:
    """Initialize the weights and bias for a model.

    Args:
        num_dimensions (int): The number of dimensions/features needed for initialization of weights and bias.

    Returns:
        np.ndarray: The initialized weights.
        np.ndarray: The initialized bias value.
    """
    limit = 1 / np.sqrt(num_dimensions)
    weights = np.random.uniform(-limit, limit, (num_dimensions,))
    bias = np.zeros(1)
    return weights, bias


def generate_minibatches(
    X: np.ndarray, y: np.ndarray, batch_size: int, shuffle: bool = False
) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
    """Generate batches of the data based on the batch_size.
    This method also handles cases where the number of samples is not divisible by the batch size.
    Example:
        X.shape -> (50, 7)
        y.shape -> (50, 1)
        batch_size -> 15
        generate_minibatches(X, y, batch_size) -> batches returned of shapes (for X) -> (15, 7), (15, 7), (15, 7), (5, 7)

    Args:
        X (np.ndarray): The input data.
        y (np.ndarray): The true labels of the data.
        batch_size (int): An integer which determines the size of each batch (number of samples in each batch).
        shuffle (bool, optional): A flag which determines if the training set should be shuffled before batches are created. Defaults to False.

    Yields:
        np.ndarray: Batches for X of size batch_size (same np.adarray format as X and y).
        np.ndarray: Batches for y of size batch_size (same np.adarray format as X and y).
    """
    if X.shape[0] != y.shape[0]:
        raise InvalidShapeError("X and y must have the same number of samples.")

    if shuffle:
        indices = np.arange(X.shape[0])
        np.random.shuffle(indices)
    for start_idx in range(0, X.shape[0], batch_size):
        end_idx = min(start_idx + batch_size, X.shape[0])
        if shuffle:
            batch = indices[start_idx:end_idx]  # type: ignore
        else:
            batch = slice(start_idx, end_idx)  # type: ignore
        yield X[batch], y[batch]



from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Tuple

import numpy as np
from showml.losses.base_loss import Loss
from showml.optimizers.base_optimizer import Optimizer
from showml.utils.dataset import Dataset
from showml.utils.model_utilities import generate_minibatches, initialize_params
from showml.utils.plots import generic_metric_plot
from showml.deep_learning.activations import Sigmoid


class Regression(ABC):
    """Base Regression class.
    """

    def compile(
        self, optimizer: Optimizer, loss: Loss, metrics: List[Callable] = []
    ) -> None:
        """Compiles the model with the specified optimizer and evaluation metrics.
        This method also initializes the model.history object to store metric values during training.

        Args:
            optimizer (Optimizer): The optimizer to be used for training (showml.optimizers).
            loss (Loss): The loss function used by the model to evaluate the solution.
            metrics (List[Callable], optional): A list of metrics which have to be calculated and displayed for model evaluation. Defaults to [].
        """
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics
        self.history: Dict[str, List[float]] = {
            metric.__name__: [] for metric in self.metrics
        }

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Computes a forward pass of the model on the given data.

        Args:
            X (np.ndarray): The input data to the network.

        Returns:
            np.ndarray: Outputs of the last layer of the network [shape: (num_samples_of_X x num_classes)]].
        """
        pass

    def evaluate(self, X: np.ndarray, y: np.ndarray) -> None:
        """Evaluate the model and display all the required metrics (accuracy, r^2 score, etc.).

        Args:
            X (np.ndarray): The input dataset.
            y (np.ndarray): The true labels of the training data.
        """
        z = self.predict(X)

        for metric in self.metrics:
            self.history[metric.__name__].append(metric(y, z))

        text_to_display = ""
        for metric_name in self.history:
            text_to_display += f", {metric_name}: {self.history[metric_name][-1]}"
        print(text_to_display)

    def plot_metrics(self) -> None:
        """Display the plot after training for the specified metrics
        """
        for metric in self.history:
            generic_metric_plot(metric, self.history[metric])

    def optimize(
        self, X: np.ndarray, y: np.ndarray, z: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """This method optimizes the weights and bias of the model using the specified loss function and optimizer.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The true labels/values.
            z (np.ndarray): The predicted labels/values.

        Returns:
            np.ndarray: The Updated weights.
            np.ndarray: The Updated bias value(s).
        """
        dw, db = self.loss.parameter_gradient(X, y, z)
        weights, bias = self.optimizer.update_weights(self.weights, self.bias, dw, db)
        return weights, bias

    def fit(self, dataset: Dataset, batch_size: int = 32, epochs: int = 1) -> None:
        """This method trains the model given the showml.utils.dataset.Dataset object (initialized with input data X and labels y).

        Args:
            dataset (Dataset): An object of the  showml.utils.dataset.Dataset class - the input dataset and true labels/values of the dataset.
            batch_size (int, optional): Number of samples per gradient update. Defaults to 32.
            epochs (int, optional): The number of epochs for training. Defaults to 50.
        """
        num_samples, num_dimensions = dataset.X.shape
        self.weights, self.bias = initialize_params(num_dimensions)

        for epoch in range(1, epochs + 1):
            print(f"Epoch: {epoch}/{epochs}", end="")

            for X_batch, y_batch in generate_minibatches(
                dataset.X, dataset.y, batch_size, shuffle=True
            ):
                # Forward pass
                z = self.predict(X_batch)

                # Optimize weights
                self.weights, self.bias = self.optimize(X_batch, y_batch, z)

            # Evaluate the model on the entire dataset
            self.evaluate(dataset.X, dataset.y)


class LinearRegression(Regression):
    def predict(self, X: np.ndarray) -> np.ndarray:
        return np.dot(X, self.weights) + self.bias


class LogisticRegression(Regression):
    def sigmoid(self, X: np.ndarray) -> np.ndarray:
        """The Sigmoid activation function.

        Args:
            X (np.ndarray): The input to the sigmoid function.

        Returns:
            np.ndarray: The output after passing the input through a sigmoid function (showml.deep_learning.activations.Sigmoid).
        """
        return Sigmoid().forward(X)

    def predict(self, X: np.ndarray) -> np.ndarray:
        return self.sigmoid(np.dot(X, self.weights) + self.bias)



import numpy as np


def normalize(X: np.ndarray) -> np.ndarray:
    """Normalize the array along the column by subracting the mean and dividing by the standard deviation.

    Args:
        X (np.ndarray): The array to be normalized.

    Returns:
        np.ndarray: The normalized array.
    """
    return (X - np.mean(X, axis=0)) / np.std(X, axis=0)


def one_hot_encode(X: np.ndarray) -> np.ndarray:
    """This method performs one hot encoding on a NumPy array.

    Args:
        X (np.ndarray): The input array to be encoded.

    Returns:
        np.ndarray: A one hot encoded array.
    """
    one_hot_encoded = np.zeros((X.size, X.max() + 1))
    one_hot_encoded[np.arange(X.size), X] = 1
    return one_hot_encoded



from showml.simulations.conways_game_of_life.event import Action
from showml.simulations.conways_game_of_life.game_window import GameWindow
from showml.simulations.conways_game_of_life.grid import Grid
from showml.simulations.conways_game_of_life.event_handler import EventHandler


class GameOfLife:
    """A simulation of Conway's Game of Life (Cellular Automaton): https://en.wikipedia.org/wiki/Conway%27s_Game_of_Life
    """

    def __init__(self, num_rows: int = 50, num_columns: int = 100) -> None:
        """Constructor for the GameOfLife class.

        Args:
            num_rows (int, optional): Number of rows in the grid. Defaults to 50.
            num_columns (int, optional): Number of columns in the grid. Defaults to 100.
        """
        self.window = GameWindow(Grid(num_rows, num_columns))
        self.event_handler = EventHandler(self.window)
        self.delay: int = 0
        self.game_running: bool = False
        self.event_dict = {
            Action.START: self.start_event,
            Action.STOP: self.stop_event,
            Action.RESET: self.reset_event,
            Action.CELL_TOGGLE: self.cell_toggle_event,
            Action.NO_EVENT: self.no_event,
        }

    def start_event(self, row: int, column: int) -> None:
        """This method runs when the game starts.
        """
        self.game_running = True
        self.delay = 150

    def stop_event(self, row: int, column: int) -> None:
        """This method runs when the game is stopped.
        """
        self.game_running = False
        self.delay = 0

    def reset_event(self, row: int, column: int) -> None:
        """This method runs when the RESET button is pressed.
        """
        self.game_running = False
        self.window.grid.reset_grid()
        self.delay = 0

    def cell_toggle_event(self, row: int, column: int) -> None:
        """This method runs when a cell value is toggled by clicking on the cell in the grid.

        Args:
            row (int): The row index of the cell which has been toggled.
            column (int): The column index of the cell which has been toggled.
        """
        self.window.grid.toggle_cell_value(row, column)

    def no_event(self, row: int, column: int) -> None:
        pass

    def run(self) -> None:
        """This method runs the game loop by communicating with the EventHandler to receive event information.
        """
        while True:
            if self.game_running:
                self.window.grid.update_grid()

            event = self.event_handler.get_event()

            # Call the event function
            self.event_dict[event.action](event.row, event.column)

            self.window.display_window_and_grid(self.delay)



from showml.losses import MeanSquaredError
import numpy as np
from numpy.testing import assert_almost_equal


def test_mse_objective_simple() -> None:
    y = np.array([3, -0.5, 2, 7])
    z = np.array([2.5, 0.0, 2, 8])
    MSE = MeanSquaredError()
    assert MSE.objective(y, z) == 0.375


def test_mse_objective_complex() -> None:
    y = np.array([0.5, 1, -1, 1, 7, -6])
    z = np.array([0, 2, -1, 2, 8, -5])
    MSE = MeanSquaredError()
    assert MSE.objective(y, z) == 0.7083333333333334


def test_mse_objective_no_error() -> None:
    y = np.array([0, 1, 3, 2, 5.78])
    z = np.array([0, 1, 3, 2, 5.78])
    MSE = MeanSquaredError()
    assert MSE.objective(y, z) == 0.0


def test_mse_gradient() -> None:
    X = np.array([[0.5, 1], [-1, 1], [7, -6], [6, 8], [1, 1]])
    y = np.array([0.5, 1, -1, 1, -6])
    z = np.array([0, 2, -1, 2, -5])
    MSE = MeanSquaredError()
    dw, db = MSE.parameter_gradient(X, y, z)
    assert_almost_equal(dw, [1.15, 1.9])
    assert db == 0.5


def test_mse_gradient_zero() -> None:
    X = np.array([[0.5, 1], [-1, 1], [7, -6], [6, 8], [1, 1]])
    y = np.array([0, 1, 3, 2, 5.78])
    z = np.array([0, 1, 3, 2, 5.78])
    MSE = MeanSquaredError()
    dw, db = MSE.parameter_gradient(X, y, z)
    assert_almost_equal(dw, [0.0, 0.0])
    assert db == 0.0



from setuptools import setup, find_packages
import os

repository_dir = os.path.dirname(__file__)

with open(os.path.join(repository_dir, "README.md")) as fh:
    long_description = fh.read()

setup(
    name="showml",
    version="1.5.7",
    packages=find_packages(exclude="tests"),
    description="A Python package of Machine Learning Algorithms implemented from scratch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hasnainroopawalla/ShowML",
    author="Hasnain Roopawalla",
    author_email="hasnain.roopawalla@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="python, machinelearning, ai",
    python_requires=">=3.6",
)



from typing import Any, Dict, Tuple

import numpy as np
from showml.optimizers.base_optimizer import Optimizer
from showml.utils.exceptions import InvalidValueError


class SGD(Optimizer):
    def __init__(self, learning_rate: float = 0.001, momentum: float = 0.0):
        """The Stochastic Gradient Descent Optimizer (SGD).

        Args:
            momentum (float, optional): This parameter builds inertia to overcome noisy gradients. Previous gradient values are stored and used to update the weights and bias at the current step. Defaults to 0.0.
        """
        if momentum < 0.0 or momentum > 1.0:
            raise InvalidValueError("Momentum value must be between 0-1")

        self.momentum = momentum
        self.prev_change: Dict[str, Any] = {"weights": np.array([]), "bias": 0.0}
        super().__init__(learning_rate)

    def update_weights(
        self, weights: np.ndarray, bias: np.ndarray, dw: np.ndarray, db: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.prev_change["weights"].shape[0] == 0:
            self.prev_change["weights"] = np.zeros(np.shape(weights))
            self.prev_change["bias"] = np.zeros(np.shape(bias))

        curr_change_weights = self.learning_rate * dw + (
            self.momentum * self.prev_change["weights"]
        )
        curr_change_bias = self.learning_rate * db + (
            self.momentum * self.prev_change["bias"]
        )

        weights -= curr_change_weights
        bias -= curr_change_bias

        # Store the current gradient for the next iteration
        self.prev_change["weights"] = curr_change_weights
        self.prev_change["bias"] = curr_change_bias

        return weights, bias


class AdaGrad(Optimizer):
    def __init__(self, learning_rate: float = 0.01, epsilon: float = 1e-8):
        """The Adaptive Gradient Descent Optimizer.
        Reference: https://ruder.io/optimizing-gradient-descent/index.html#adagrad

        Args:
            epsilon (float, optional): A smoothing term that ensures the denominator is > 0 (to avoid division by zero). Defaults to 1e-8.
        """
        self.G: Dict[str, Any] = {"weights": np.array([]), "bias": 0.0}
        self.epsilon = epsilon
        super().__init__(learning_rate)

    def update_weights(
        self, weights: np.ndarray, bias: np.ndarray, dw: np.ndarray, db: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:

        if self.G["weights"].shape[0] == 0:
            self.G["weights"] = np.zeros(np.shape(weights))
            self.G["bias"] = np.zeros(np.shape(bias))

        self.G["weights"] += np.square(dw)
        self.G["bias"] += np.square(db)

        weights -= self.learning_rate * (dw / np.sqrt(self.G["weights"] + self.epsilon))
        bias -= self.learning_rate * (db / np.sqrt(self.G["bias"] + self.epsilon))

        return weights, bias


class RMSProp(Optimizer):
    def __init__(
        self, learning_rate: float = 0.001, rho: float = 0.9, epsilon: float = 1e-8
    ):
        """The Root Mean Squared Propagation (RMSProp) Optimizer.
        Reference: https://machinelearningmastery.com/gradient-descent-with-rmsprop-from-scratch/

        Args:
            rho (float, optional): A parameter to determine the weight to be given to recent gradients as compared to older gradients. Defaults to 0.9.
            epsilon (float, optional): A smoothing term that ensures the denominator is > 0 (to avoid division by zero). Defaults to 1e-8.
        """
        self.G: Dict[str, Any] = {"weights": np.array([]), "bias": 0.0}
        self.rho = rho
        self.epsilon = epsilon
        super().__init__(learning_rate)

    def update_weights(
        self, weights: np.ndarray, bias: np.ndarray, dw: np.ndarray, db: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:

        if self.G["weights"].shape[0] == 0:
            self.G["weights"] = np.zeros(np.shape(weights))
            self.G["bias"] = np.zeros(np.shape(bias))

        self.G["weights"] = (self.rho * self.G["weights"]) + (
            (1 - self.rho) + np.square(dw)
        )
        self.G["bias"] = (self.rho * self.G["bias"]) + ((1 - self.rho) + np.square(db))

        weights -= self.learning_rate * (dw / np.sqrt(self.G["weights"] + self.epsilon))
        bias -= self.learning_rate * (db / np.sqrt(self.G["bias"] + self.epsilon))

        return weights, bias



from showml.losses.metrics import r2_score
import numpy as np


def test_r2_score_simple() -> None:
    y = np.array([4, 3, 2, 1])
    z = np.array([1, 2, 3, 4])
    assert r2_score(y, z) == -3.0


def test_r2_score_float_1D() -> None:
    y = np.array([3, -0.5, 2, 7])
    z = np.array([2.5, 0.0, 2, 8])
    assert r2_score(y, z) == 0.9486081370449679


def test_r2_score_float_2D() -> None:
    z = np.array([[0, 2], [-1, 2], [8, -5]])
    y = np.array([[0.5, 1], [-1, 1], [7, -6]])
    assert r2_score(y, z) == 0.9512661251791686
